import os
import webbrowser
import tempfile


GLOBAL = {'VERBOSE': True, 
          'DASK_CLIENT': None,
          'CACHE':None}


def cache(cachedir=None):
    if cachedir is not None:
        os.environ['VIPY_CACHE'] = cachedir
    return os.environ['VIPY_CACHE'] if 'VIPY_CACHE' in os.environ else None
    

class Dask(object):
    def __init__(self, num_processes, dashboard=False):
        assert isinstance(num_processes, int) and num_processes >=2, "num_processes must be >= 2"

        from vipy.util import try_import
        try_import('dask', 'dask distributed')
        import dask
        from dask.distributed import Client
        from dask.distributed import as_completed, wait
        from dask.config import set as dask_config_set
        
        dask_config_set({"distributed.comm.timeouts.tcp": "50s"})
        dask_config_set({"distributed.comm.timeouts.connect": "10s"})        
        self._num_processes = num_processes
        self._client = Client(name='vipy', 
                              scheduler_port=0, 
                              dashboard_address=None if not dashboard else ':0',  # random port
                              processes=True, 
                              threads_per_worker=1, 
                              n_workers=num_processes, 
                              env={'VIPY_BACKEND':'Agg'},
                              direct_to_workers=True,
                              local_directory=tempfile.mkdtemp())

    def __repr__(self):
        return str('<vipy.globals.dask: num_processes=%d%s>' % (self._num_processes, '' if self._num_processes==0 or len(self._client.dashboard_link)==0 else ', dashboard="%s"' % str(self._client.dashboard_link)))

    def dashboard(self):        
        webbrowser.open(self._client.dashboard_link) if len(self._client.dashboard_link)>0 else None
    
    def num_processes(self):
        return self._num_processes

    def shutdown(self):
        self._client.close()
        self._num_processes = 0
        GLOBAL['DASK_CLIENT'] = None
        return self

    def client(self):
        return self._client

    
def dask(num_processes=None, dashboard=False):
    if GLOBAL['DASK_CLIENT'] is None and num_processes is not None:
        GLOBAL['DASK_CLIENT'] = Dask(num_processes, dashboard=dashboard)        
    elif GLOBAL['DASK_CLIENT'] is not None and num_processes is not None and GLOBAL['DASK_CLIENT'].num_processes() != num_processes:
        GLOBAL['DASK_CLIENT'].shutdown()
        GLOBAL['DASK_CLIENT'] = Dask(num_processes, dashboard=dashboard)        
    return GLOBAL['DASK_CLIENT']


def num_workers(n=None):
    if n is not None:
        return dask(num_processes=n)
    return 1 if dask() is None else dask().num_processes()

def max_workers():
    import multiprocessing
    return dask(num_processes=multiprocessing.cpu_count())



import numpy as np
import scipy.ndimage


def signsqrt(x):
    """Return the signed square root of elements in x"""
    return np.multiply(np.sign(x), np.sqrt(np.abs(x)))


def runningmean(X, n):
    """Compute the running unweighted mean of X row-wise, with a history of n, with reflection along each column"""
    return scipy.ndimage.uniform_filter1d(X.astype(np.float32), axis=0, size=n, mode='reflect')


def find_closest_positive_divisor(a, b):
    """Return non-trivial positive integer divisor (bh) of (a) closest to (b) in abs(b-bh) such that a % bh == 0.  This uses exhaustive search, which is inefficient for large a."""
    assert a>0 and b>0
    if a<=b:
        return a
    for k in range(0, a-b+1):
        bh = b + k
        if bh>1 and a % bh == 0:
            return bh
        bh = b - k
        if bh>1 and a % bh == 0:
            return bh
    return a  # should never get here, since bh=a is always a solution



import os
from vipy.util import filetail, remkdir, readjson
import vipy.downloader
from vipy.video import VideoCategory, Video
import numpy as np


# http://activity-net.org/download.html
URL = 'http://ec2-52-25-205-214.us-west-2.compute.amazonaws.com/files/activity_net.v1-3.min.json'


class ActivityNet(object):
    def __init__(self, datadir):
        """Activitynet, provide a datadir='/path/to/store/activitynet' """
        self.datadir = remkdir(datadir)

    def __repr__(self):
        return str('<vipy.dataset.activitynet: "%s">' % self.datadir)

    def download(self):
        vipy.downloader.download(URL, os.path.join(self.datadir, filetail(URL)))
        return self

    def _isdownloaded(self):
        return os.path.exists(os.path.join(self.datadir, 'activity_net.v1-3.min.json'))
    
    def _dataset(self, subset):
        assert self._isdownloaded(), "Dataset not downloaded.  download() first or manually download '%s' into '%s'" % (self._url, self.datadir)        
        jsonfile = os.path.join(self.datadir, filetail(URL))
        json = readjson(jsonfile)
        return [VideoCategory(url=v['url'],
                              filename=os.path.join(self.datadir, youtubeid),
                              category=a['label'],
                              startsec=float(a['segment'][0]),
                              endsec=float(a['segment'][1]))
                for (youtubeid, v) in json['database'].items()
                for a in v['annotations']
                if v['subset'] == subset]

    def trainset(self):
        return self._dataset('training')

    def testset(self):
        """ActivityNet test set does not include any annotations"""
        assert self._isdownloaded(), "Dataset not downloaded.  download() first or manually download '%s' into '%s'" % (self._url, self.datadir)        
        json = readjson(os.path.join(self.datadir, filetail(URL)))
        return [Video(url=v['url'], filename=os.path.join(self.datadir, youtubeid)) for (youtubeid, v) in json['database'].items() if v['subset'] == 'testing']

    def valset(self):
        return self._dataset('validation')
    
    



import numpy as np
from vipy.geometry import BoundingBox
from vipy.util import isstring, tolist
import uuid
import copy


class Detection(BoundingBox):
    """vipy.object.Detection class
    
    This class represent a single object detection in the form a bounding box with a label and confidence.
    The constructor of this class follows a subset of the constructor patterns of vipy.geometry.BoundingBox

    >>> d = vipy.object.Detection(category='Person', xmin=0, ymin=0, width=50, height=100)
    >>> d = vipy.object.Detection(label='Person', xmin=0, ymin=0, width=50, height=100)  # "label" is an alias for "category"
    >>> d = vipy.object.Detection(label='John Doe', shortlabel='Person', xmin=0, ymin=0, width=50, height=100)  # shortlabel is displayed
    >>> d = vipy.object.Detection(label='Person', xywh=[0,0,50,100])

    """

    def __init__(self, label=None, xmin=None, ymin=None, width=None, height=None, xmax=None, ymax=None, confidence=None, xcentroid=None, ycentroid=None, category=None, xywh=None, shortlabel=None, attributes=None):
        super(Detection, self).__init__(xmin=xmin, ymin=ymin, width=width, height=height, xmax=xmax, ymax=ymax, xcentroid=xcentroid, ycentroid=ycentroid, xywh=xywh)
        assert not (label is not None and category is not None), "Constructor requires either label or category kwargs, not both"
        self._id = uuid.uuid1().hex        
        self._label = category if category is not None else label
        self._shortlabel = self._label if shortlabel is None else shortlabel
        self._confidence = float(confidence) if confidence is not None else confidence
        self.attributes = attributes if attributes is not None else {}

    def __repr__(self):
        strlist = []
        if self.category() is not None:
            strlist.append('category="%s"' % self.category())
        if self.isvalid():
            strlist.append('bbox=(xmin=%1.1f, ymin=%1.1f, width=%1.1f, height=%1.1f)' %
                           (self.xmin(), self.ymin(),self.width(), self.height()))
        if self._confidence is not None:
            strlist.append('conf=%1.3f')
        return str('<vipy.object.detection: %s>' % (', '.join(strlist)))

    def __eq__(self, other):
        """Detection equality when bounding boxes and categories are equivalent"""
        return isinstance(other, Detection) and self.xywh() == other.xywh() and self.category() == other.category()

    def __str__(self):
        return self.__repr__()

    def dict(self):
        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel() ,'boundingbox':super(Detection, self).dict(),
                'attributes':self.attributes,  # these may be arbitrary user defined objects
                'confidence':self._confidence}

    def nocategory(self):
        self._label = None
        return self

    def category(self, category=None):
        """Update the category of the detection"""
        if category is None:
            return self._label
        else:
            self._label = category
            return self

    def shortlabel(self, label=None):
        """A optional shorter label string to show in the visualizations, defaults to category()"""        
        if label is not None:
            self._shortlabel = label
            return self
        else:
            return self._shortlabel

    def label(self, label):
        """Alias for category"""
        return self.category(label)

    def id(self):
        return self._id

    def clone(self):
        return copy.deepcopy(self)
    

class Track(object):
    """vipy.object.Track class
    
    A track represents one or more labeled bounding boxes of an object instance through time.  A track is defined as a finite set of labeled boxes observed 
    at keyframes, which are discrete observations of this instance.  Each keyframe has an associated vipy.geometry.BoundingBox() which defines the spatial bounding box
    of the instance in this keyframe.  The kwarg "interpolation" defines how the track is interpolated between keyframes, and the kwarg "boundary" defines how the 
    track is interpolated outside the (min,max) of the keyframes.  

    Valid constructors are:

    >>> t = vipy.object.Track(keyframes=[0,100], boxes=[vipy.geometry.BoundingBox(0,0,10,10), vipy.geometry.BoundingBox(0,0,20,20)], label='Person')
    >>> t = vipy.object.Track(keyframes=[0,100], boxes=[vipy.geometry.BoundingBox(0,0,10,10), vipy.geometry.BoundingBox(0,0,20,20)], label='Person', interpolation='linear')
    >>> t = vipy.object.Track(keyframes=[10,100], boxes=[vipy.geometry.BoundingBox(0,0,10,10), vipy.geometry.BoundingBox(0,0,20,20)], label='Person', boundary='strict')

    Tracks can be constructed incrementally:

    >>> t = vipy.object.Track('Person')
    >>> t.add(0, vipy.geometry.BoundingBox(0,0,10,10))
    >>> t.add(100, vipy.geometry.BoundingBox(0,0,20,20))

    Tracks can be resampled at a new framerate, as long as the framerate is known when the keyframes are extracted

    >>> t.framerate(newfps)

    """

    def __init__(self, keyframes, boxes, category=None, label=None, confidence=None, framerate=None, interpolation='linear', boundary='strict', shortlabel=None, attributes=None):

        keyframes = tolist(keyframes)
        boxes = tolist(boxes)        
        assert isinstance(keyframes, tuple) or isinstance(keyframes, list), "Keyframes are required and must be tuple or list"
        assert isinstance(boxes, tuple) or isinstance(boxes, list), "Keyframe boundingboxes are required and must be tuple or list"
        assert all([isinstance(bb, BoundingBox) for bb in boxes]), "Keyframe bounding boxes must be vipy.geometry.BoundingBox objects"
        assert all([bb.isvalid() for bb in boxes]), "All keyframe bounding boxes must be valid"        
        assert not (label is not None and category is not None), "Constructor requires either label or category kwargs, not both"                
        assert len(keyframes) == len(boxes), "Boxes and keyframes must be the same length, there must be a one to one mapping of frames to boxes"
        assert boundary in set(['extend', 'strict']), "Invalid interpolation boundary - Must be ['extend', 'strict']"
        assert interpolation in set(['linear']), "Invalid interpolation - Must be ['linear']"
                
        self._id = uuid.uuid1().hex
        self._label = category if category is not None else label
        self._shortlabel = self._label if shortlabel is None else shortlabel
        self._framerate = framerate
        self._interpolation = interpolation
        self._boundary = boundary
        self.attributes = attributes if attributes is not None else {}        
        self._keyframes = keyframes
        self._keyboxes = boxes
        
        # Sorted increasing frame order
        if len(keyframes) > 0 and len(boxes) > 0:
            (keyframes, boxes) = zip(*sorted([(f,bb) for (f,bb) in zip(keyframes, boxes)], key=lambda x: x[0]))
            self._keyframes = list(keyframes)
            self._keyboxes = list(boxes)
        
    def __repr__(self):
        strlist = []
        if self.category() is not None:
            strlist.append('category="%s"' % self.category())
        if self.endframe() - self.startframe() > 0:
            strlist.append('startframe=%d, endframe=%d' % (self.startframe(), self.endframe()))
        strlist.append('keyframes=%d' % len(self._keyframes))
        return str('<vipy.object.track: %s>' % (', '.join(strlist)))

    def __getitem__(self, k):
        """Interpolate the track at frame k"""
        return self._linear_interpolation(k)

    def __iter__(self):
        """Iterate over the track interpolating each frame from min(keyframes) to max(keyframes)"""
        for k in range(self.startframe(), self.endframe()+1):
            yield self._linear_interpolation(k)

    def __len__(self):
        """The length of a track is the total number of interpolated frames"""
        return self.endframe() - self.startframe() + 1

    def dict(self):
        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel(), 'keyframes':self._keyframes, 'framerate':self._framerate, 
                'boundingbox':[bb.dict() for bb in self._keyboxes], 'attributes':self.attributes}

    def add(self, keyframe, box):
        """Add a new keyframe and associated box to track, preserve sorted order of keyframes"""
        assert isinstance(box, BoundingBox), "Invalid input - Box must be vipy.geometry.BoundingBox()"
        assert box.isvalid(), "Invalid input - Box must be non-degenerate"
        self._keyframes.append(keyframe)
        self._keyboxes.append(box)
        if len(self._keyframes) > 1 and keyframe < self._keyframes[-2]:
            (self._keyframes, self._keyboxes) = zip(*sorted([(f,bb) for (f,bb) in zip(self._keyframes, self._keyboxes)], key=lambda x: x[0]))        
            self._keyframes = list(self._keyframes)
            self._keyboxes = list(self._keyboxes)
        return self
        
    def keyframes(self):
        """Return keyframe frame indexes where there are track observations"""
        return self._keyframes

    def meanshape(self):
        """Return the mean (width,height) of the box during the track"""
        return np.mean([bb.shape() for bb in self._keyboxes], axis=0)
            
    def framerate(self, fps):
        """Resample keyframes from known original framerate set by constructor to be new framerate fps"""
        assert self._framerate is not None, "Framerate conversion requires that the framerate is known for current keyframes.  This must be provided to the vipy.object.Track() constructor."
        self._keyframes = [int(np.round(f*(fps/float(self._framerate)))) for f in self._keyframes]
        self._framerate = fps
        return self
        
    def startframe(self):
        return np.min(self._keyframes)

    def endframe(self):
        return np.max(self._keyframes)

    def _linear_interpolation(self, k):
        """Linear bounding box interpolation at frame=k given observed boxes (x,y,w,h) at keyframes.  
        This returns a vipy.object.Detection() which is the interpolation of the Track() at frame k
        If self._boundary='extend', then boxes are repeated if the interpolation is outside the keyframes
        If self._boundary='strict', then interpolation returns None if the interpolation is outside the keyframes
        """
        (xmin, ymin, width, height) = zip(*[bb.to_xywh() for bb in self._keyboxes])
        d = Detection(xmin=np.interp(k, self._keyframes, xmin),
                      ymin=np.interp(k, self._keyframes, ymin),
                      width=np.interp(k, self._keyframes, width),
                      height=np.interp(k, self._keyframes, height),
                      category=self.category(),
                      shortlabel=self.shortlabel())
        d.attributes['trackid'] = self.id()  # for correspondence of detections to tracks
        return d if self._boundary == 'extend' else (None if not self.during(k) else d)

    def category(self, label=None):
        if label is not None:
            self._label = label
            return self
        else:
            return self._label

    def label(self, label):
        """Alias for category"""
        return self.category(label)
        
    def shortlabel(self, label=None):
        """A optional shorter label string to show in the visualizations"""                
        if label is not None:
            self._shortlabel = label
            return self
        else:
            return self._shortlabel

    def during(self, k):
        """Is frame during the time interval (startframe, endframe) inclusive?"""        
        return k >= self.startframe() and k <= self.endframe()

    def offset(self, dt=0, dx=0, dy=0):
        self._keyboxes = [bb.offset(dx, dy) for bb in self._keyboxes]
        self._keyframes = list(np.array(self._keyframes) + dt)
        return self

    def rescale(self, s):
        """Rescale track boxes by scale factor s"""
        self._keyboxes = [bb.rescale(s) for bb in self._keyboxes]
        return self

    def scale(self, s):
        """Alias for rescale"""
        return self.rescale(s)

    def scalex(self, sx):
        """Rescale track boxes by scale factor sx"""
        self._keyboxes = [bb.scalex(sx) for bb in self._keyboxes]
        return self

    def scaley(self, sy):
        """Rescale track boxes by scale factor sx"""
        self._keyboxes = [bb.scaley(sy) for bb in self._keyboxes]
        return self

    def dilate(self, s):
        """Dilate track boxes by scale factor s"""
        self._keyboxes = [bb.dilate(s) for bb in self._keyboxes]
        return self

    def rot90cw(self, H, W):
        """Rotate an image with (H,W)=shape 90 degrees clockwise and update all boxes to be consistent"""
        self._keyboxes = [bb.rot90cw(H, W) for bb in self._keyboxes]
        return self

    def rot90ccw(self, H, W):
        """Rotate an image with (H,W)=shape 90 degrees clockwise and update all boxes to be consistent"""
        self._keyboxes = [bb.rot90ccw(H, W) for bb in self._keyboxes]
        return self

    def id(self):
        return self._id

    def clone(self):
        return copy.deepcopy(self)

    def boundingbox(self):
        """The bounding box of a track is the smallest spatial box that contains all of the detections, or None if there are no detections"""
        d = self._keyboxes[0].clone() if len(self._keyboxes) >= 1 else None
        return d.union(self._keyboxes[1:]) if (d is not None and len(self._keyboxes) >= 2) else d

    def clip(self, startframe, endframe):
        """Clip a track to be within (startframe,endframe) with strict boundary handling"""
        if self[startframe] is not None:
            self.add(startframe, self[startframe])
        if self[endframe] is not None:
            self.add(endframe, self[endframe])
        self._keyframes = [f for (f,bb) in zip(self._keyframes, self._keyboxes) if f>=startframe and f<=endframe]  # may be empty
        self._keyboxes = [bb for (f,bb) in zip(self._keyframes, self._keyboxes) if f>=startframe and f<=endframe]  # may be empty
        if len(self._keyframes) == 0 or len(self._keyboxes) == 0:
            raise ValueError('Track does not contain any keyboxes within the requested frames (%d,%d)' % (startframe, endframe))
        self._boundary = 'strict'
        return self

    def iou(self, other):
        """Compute the spatial IoU between two tracks as the mean IoU per frame in the range (self.startframe(), self.endframe())"""
        assert isinstance(other, Track), "Invalid input - must be vipy.object.Track()"
        return np.mean([self[k].iou(other[k]) if self.during(k) and other.during(k) else 0.0 for k in range(self.startframe(), self.endframe())])

    def imclip(self, width, height):
        """Clip the track to the image rectangle (width, height).  If a keybox is outside the image rectangle, remove it otherwise clip to the image rectangle. 
           This operation can change the length of the track and the size of the keyboxes.  The result may be an empty track if the track is completely outside
           the image rectangle, which results in an exception.
        """
        clipped = [(f, bb.imclip(width=width, height=height)) for (f,bb) in zip(self._keyframes, self._keyboxes) if bb.hasoverlap(width=width, height=height)]
        if len(clipped) > 0:
            (self._keyframes, self._keyboxes) = zip(*clipped)
            (self._keyframes, self._keyboxes) = (list(self._keyframes), list(self._keyboxes))
            return self
        else:
            raise ValueError('All key boxes for track outside image rectangle')


class Activity(object):
    """vipy.object.Activity class
    
    An activity is a grouping of one or more tracks involved in an activity within a given startframe and endframe.
    The activity occurs at a given (startframe, endframe), where these frame indexes are extracted at the provided framerate.
    All objects are passed by reference with a globally unique track ID, for the tracks involved with the activity.  This 
    is done since tracks can exist after an activity completes, and that tracks should update the spatial transformation of boxes.
    The shortlabel defines the string shown on the visualization video.

    Valid constructors

    >>> t = vipy.object.Track(category='Person').add(...))
    >>> a = vipy.object.Activity(startframe=0, endframe=10, category='Walking', tracks={t.id():t})

    """
    def __init__(self, startframe, endframe, framerate=None, label=None, shortlabel=None, category=None, tracks=None, attributes=None):
        assert not (label is not None and category is not None), "Activity() Constructor requires either label or category kwargs, not both"
        assert startframe < endframe, "Start frame must be strictly less than end frame"
        assert tracks is None or isinstance(tracks, dict), "Tracks must be a dictionary {trackid:vipy.object.Track()}"
        assert tracks is None or all([isstring(k) for (k,v) in tracks.items()]) and all([isinstance(v, Track) for (k,v) in tracks.items()]), "Invalid tracks - Must be a dictionary of {str(trackid):vipy.object.Track()}"        
        assert tracks is None or all([t.during(startframe) or t.during(endframe) for t in tracks.values()]), "All tracks must be be present when this activity occurs"
    
        self._id = uuid.uuid1().hex
        self._startframe = startframe
        self._endframe = endframe
        self._framerate = framerate
        self._label = category if category is not None else label        
        self._shortlabel = self._label if shortlabel is None else shortlabel
        self._tracks = tracks if tracks is not None else {}

        self.attributes = attributes if attributes is not None else {}            
        
    def __len__(self):
        """Return activity length in frames"""
        return self.endframe() - self.startframe()

    def __getitem__(self, k):
        """Return a list Detection() objects interpolated at frame k if during activity, otherwise return None"""
        return [t[k] for (i,t) in self._tracks.items() if t[k] is not None] if self.during(k) else None
    
    def __repr__(self):
        return str('<vipy.activity: category="%s", frames=(%d,%d), tracks=%s>' % (self.category(), self.startframe(), self.endframe(), len(self.tracks())))

    def dict(self):
        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel(), 'startframe':self._startframe, 'endframe':self._endframe, 'attributes':self.attributes, 'framerate':self._framerate,
                'tracks':[t.dict() for (k,t) in self._tracks.items()]}
    
    def startframe(self):
        return self._startframe
    
    def endframe(self):
        return self._endframe

    def middleframe(self):
        return int(np.round((self.endframe() - self.startframe()) / 2.0)) + self.startframe()

    def framerate(self, fps):
        """Resample (startframe, endframe) from known original framerate set by constructor to be new framerate fps"""        
        assert self._framerate is not None, "Framerate conversion requires that the framerate is known for current activities.  This must be provided to the vipy.object.Activity() constructor."
        (self._startframe, self._endframe) = [int(np.round(f*(fps/float(self._framerate)))) for f in (self._startframe, self._endframe)]
        self._framerate = fps
        return self
    
    def category(self, label=None):
        if label is not None:
            self._label = label
            return self
        else:
            return self._label

    def label(self, label):
        """Alias for category"""
        return self.category(label)

    def categories(self):
        """Return a set of categories for the activities and objects in this activity"""
        return set([self.category()] + [t.category() for t in self.tracks().values()])
        
    def shortlabel(self, label=None):
        """A optional shorter label string to show in the visualizations"""                
        if label is not None:
            self._shortlabel = label
            return self
        else:
            return self._shortlabel

    def add(self, track):
        """Add the track by reference to the track list for this activity, so that if the track is changed externally it is reflected here"""
        assert isinstance(track, Track), "Invalid input - must be vipy.object.Track"
        assert track.during(self.startframe()) or track.during(self.endframe()) or track.during(self.middleframe()), "The track must be present during the activity"
        self._tracks[track.id()] = track
        return self
        
    def tracks(self, tracks=None):
        """Returns a track dictionary, tracks are referenced in the dictionary and are mutable"""
        if tracks is not None:
            assert isinstance(tracks, dict), "Tracks must be dictionary of {str(trackid):vipy.object.Track()}"
            self._tracks = tracks
            return self
        else:
            return self._tracks

    def hastrack(self, track):
        """Is the track part of the activity?"""
        assert isstring(track) or isinstance(track, Track), "Invalid input - Must be a vipy.object.Track().id() or vipy.object.Track()"
        trackid = track.id() if isinstance(track, Track) else track
        return any([tid == trackid for tid in self._tracks.keys()])
            
    def during(self, frame):
        """Is frame during the time interval (startframe, endframe) inclusive?"""
        return int(frame) >= self._startframe and int(frame) <= self._endframe

    def spatial_iou(self, other):
        """Return the mean spatial intersection over union of two activities as the mean spatial IoU for the union of tracks at each frame during (startframe, endframe)
           Note that we cannot do the IoU of individual tracks because there is no way to correspond tracks within an activity, since an activity may have more than one 
           track with the same category.  
        """
        assert isinstance(other, Activity), "Invalid input - must be vipy.object.Activity()"
        return np.mean([bbi.union(self[k]).iou(bbj.union(other[k])) if (bbj is not None and bbi is not None) else 0.0 for k in range(self.startframe(), self.endframe()) for bbi in self[k] for bbj in other[k]])

    def temporal_iou(self, other):
        """Return the temporal intersection over union of two activities"""
        assert isinstance(other, Activity), "Invalid input - must be vipy.object.Activity()"
        t_start = min(self.startframe(), other.startframe())
        t_end = max(self.endframe(), other.endframe())
        t_union = float(t_end - t_start)
        
        t_start = max(self.startframe(), other.startframe())
        t_end = min(self.endframe(), other.endframe())
        t_intersection = float(t_end - t_start)
        
        return (t_intersection / t_union) if t_intersection > 0 else 0
    
    def offset(self, dt):
        self._startframe = self._startframe + dt
        self._endframe = self._endframe + dt
        self._tracks = {ti:t.offset(dt=dt) for (ti,t) in self._tracks.items()}
        return self
    
    def id(self):
        return self._id

    def boundingbox(self):
        """The bounding box of an activity is the smallest bounding box for all tracks in the activity (inclusive of start and endframes), or None of there are no boxes""" 
        boxes = [t.clone().clip(self.startframe(), self.endframe()+1).boundingbox() for (i,t) in self.tracks().items()]
        return boxes[0].clone().union(boxes[1:]) if len(boxes)>0 else None

    def imagebox(self, width, height):
        """The image box of an activity is the smallest bounding box for all tracks in the activity (inclusive of start and endframes) that is within the image rectangle (width, height), or None if there are no boxes""" 
        bb = self.boundingbox()
        return None if bb is None or not bb.hasoverlap(width=width, height=height) else bb.imclipshape(width, height)

    def clone(self):
        return copy.deepcopy(self)
    



import os
from vipy.util import remkdir, readjson, readyaml, findyaml, findvideo, filetail, findjson, filebase, readlist, groupbyasdict, save, flatlist, isstring, tempdir, readcsv, delpath
from vipy.video import VideoCategory, Scene
from vipy.object import Track, Activity
from vipy.geometry import BoundingBox
from vipy.batch import Batch
from vipy.show import colorlist
import vipy.metrics
import numpy as np
import warnings
import shutil
import vipy.globals


class KF1(object):
    def __init__(self, videodir, repodir, contrib=False, stride=20, verbose=True, n_videos=None, d_category_to_shortlabel=None):
        """Parse MEVA annotations (http://mevadata.org) for KNown Facility 1 dataset into vipy.video.Scene() objects
       
        Kwiver packet format: https://gitlab.kitware.com/meva/meva-data-repo/blob/master/documents/KPF-specification-v4.pdf
        Inputs:
          -videodir=str:  path to Directory containing 'drop-01' 
          -repodir=str:  path to directory containing clone of https://gitlab.kitware.com/meva/meva-data-repo
          -stride=int: the temporal stride in frames for importing bounding boxes, vipy will do linear interpoluation and boundary handling
          -n_videos=int:  only return an integer number of videos, useful for debugging or for previewing dataset
          -contrib=bool:  include the noisy contrib anntations from DIVA performers
          -d_category_to_shortlabel is a dictionary mapping category names to a short displayed label on the video.  The standard for visualization is that 
            tracked objects are displayed with their category label (e.g. 'Person', 'Vehicle'), and activities are labeled according to the set of objects that
            performing the activity.  When an activity occurs, the set of objects are labeled with the same color as 'Noun Verbing' (e.g. 'Person Entering', 
            'Person Reading', 'Vehicle Starting') where 'Verbing' is provided by the shortlabel.   This is optional, and will use the default mapping if None
          -verbose=bool:  Parsing verbosity
        """
        
        self.videodir = videodir
        self.repodir = repodir

        assert os.path.exists(os.path.join(self.videodir, 'drop-01')), "Invalid input - videodir must contain the drop-01, drop-02 and drop-03 subdirectories.  See http://mevadata.org/#getting-data"
        assert os.path.exists(os.path.join(self.repodir, 'annotation')), "Invalid input - repodir must contain the clone of https://gitlab.kitware.com/meva/meva-data-repo"

        self._d_category_to_shortlabel = {'person_abandons_package':'Abandoning',
                                          'person_closes_facility_door':'Closing',
                                          'person_closes_trunk':'Closing trunk',
                                          'person_closes_vehicle_door':'Closing door',
                                          'person_embraces_person':'Hugging',
                                          'person_enters_scene_through_structure':'Entering',
                                          'person_enters_vehicle':'Entering',
                                          'person_exits_scene_through_structure':'Exiting',
                                          'person_exits_vehicle':'Exiting',
                                          'hand_interacts_with_person':'Using hand',
                                          'person_carries_heavy_object':'Carrying',
                                          'person_interacts_with_laptop':'Using laptop',
                                          'person_loads_vehicle':'Loading',
                                          'person_transfers_object':'Transferring',
                                          'person_opens_facility_door':'Opening door',
                                          'person_opens_trunk':'Opening trunk',
                                          'person_opens_vehicle_door':'Opening door',
                                          'person_talks_to_person':'Talking',
                                          'person_picks_up_object':'Picking up',
                                          'person_purchases':'Purchasing',
                                          'person_reads_document':'Reading',
                                          'person_rides_bicycle':'Riding',
                                          'person_puts_down_object':'Putting down',
                                          'person_sits_down':'Sitting',
                                          'person_stands_up':'Standing',
                                          'person_talks_on_phone':'Talking',
                                          'person_texts_on_phone':'Texting',
                                          'person_steals_object':'Stealing',
                                          'person_unloads_vehicle':'Unloading',
                                          'vehicle_drops_off_person':'Dropping off',
                                          'vehicle_picks_up_person':'Picking up',
                                          'vehicle_reverses':'Reversing',
                                          'vehicle_starts':'Starting',
                                          'vehicle_stops':'Stopping',
                                          'vehicle_turns_left':'Turning left',
                                          'vehicle_turns_right':'Turning right',
                                          'vehicle_makes_u_turn':'Turning around'}
        self._d_category_to_shortlabel = {k:v.lower() for (k,v) in self._d_category_to_shortlabel.items()}
        
        self._d_oldcategory_to_shortlabel = {'Closing_Trunk':'Closing',
                                             'Open_Trunk':'Opening',
                                             'Riding':'Riding',
                                             'Talking':'Talking',
                                             'person_talks_to_person':'Talking',                                          
                                             'Transport_HeavyCarry':'Carrying',
                                             'Unloading':'Unloading',
                                             'abandon_package':'Abandoning',
                                             'hand_interaction':'Using Hand',
                                             'object_transfer':'Transferring',
                                             'person_closes_facility_door':'Closing',
                                             'person_closes_vehicle_door':'Closing',
                                             'person_enters_through_structure':'Entering',
                                             'person_enters_vehicle':'Entering',
                                             'person_exits_through_structure':'Exiting',                                          
                                             'person_exits_vehicle':'Exiting',
                                             'person_laptop_interaction':'Interacting',
                                             'person_loads_vehicle':'Loading',
                                             'person_opens_facility_door':'Opening',
                                             'person_opens_vehicle_door':'Opening',
                                             'person_person_embrace':'Hugging',
                                             'person_picks_up_object':'Picking up',
                                             'person_purchasing':'Purchasing',
                                             'person_reading_document':'Reading',
                                             'person_sets_down_object':'Setting down',
                                             'person_sitting_down':'Sitting',
                                             'person_standing_up':'Standing',
                                             'person_stands_up':'Standing',                                          
                                             'specialized_talking_phone':'Talking',
                                             'specialized_texting_phone':'Texting',    
                                             'theft':'Theft',
                                             'vehicle_drops_off_person':'Dropping off',
                                             'vehicle_picks_up_person':'Picking up',
                                             'vehicle_reversing':'Reversing', 
                                             'vehicle_starting':'Starting',
                                             'vehicle_stopping':'Stopping',
                                             'vehicle_turning_left':'Turning left',
                                             'vehicle_turning_right':'Turning right',
                                             'vehicle_u_turn':'Turning around'}

        self._d_oldcategory_to_newcategory = {k:v for (k,v) in readcsv(os.path.join(self.repodir, 'documents', 'activity-name-mapping.csv'))[1:]}

        d_category_to_shortlabel = d_category_to_shortlabel if d_category_to_shortlabel is not None else self._d_category_to_shortlabel
        d_videoname_to_path = {filebase(f):f for f in self._get_videos()}
        yamlfiles = zip(self._get_types_yaml(), self._get_geom_yaml(), self._get_activities_yaml())
        yamlfiles = [y for y in yamlfiles if contrib is True or 'contrib' not in y[0]]
        yamlfiles = list(yamlfiles)[0:n_videos] if n_videos is not None else list(yamlfiles)
        if verbose:
            print('[vipy.dataset.meva.KF1]: Loading %d YAML files' % len(yamlfiles))
            if len(yamlfiles) > 100 and vipy.globals.num_workers() == 1: 
                print('[vipy.dataset.meva.KF1]: This takes a while, consider setting vipy.globals.num_workers(8) for parallel parsing')

        if vipy.globals.num_workers() > 1:
            from vipy.batch import Batch
            self._vidlist = Batch(list(yamlfiles)).map(lambda tga: self._parse_video(d_videoname_to_path, d_category_to_shortlabel, tga[0], tga[1], tga[2], stride=stride, verbose=verbose))
        else:
            self._vidlist = [self._parse_video(d_videoname_to_path, d_category_to_shortlabel, t, g, a, stride=stride, verbose=verbose) for (t,g,a) in yamlfiles]
        self._vidlist = [v for v in self._vidlist if v is not None] 
        
    def __getitem__(self, k):
        return self._vidlist[k]

    def __iter__(self):
        for v in self._vidlist:
            yield v

    def __len__(self):
        return len(self._vidlist)
                
    def __repr__(self):
        return str('<vipy.dataset.meva.KF1: videos=%d, videodir="%s", annotationdir="%s">' % (len(self), self.videodir, self.repodir))

    def _activities_to_required_objects(self):
        """Return a dictionary of activity keys to set of required objects.  This is currently wrong."""
        raise ValueError('This mapping is currently wrong in the Kitware repository')
        d = readjson(os.path.join(self.repodir, 'annotation', 'DIVA-phase-2', 'activity-index.json'))
        return {a:set([x.replace('Construction_Vehicle', 'Vehicle') for x in d[a]['objectTypes']]) for a in self.activities()}        

    def _get_activities_yaml(self):
        return sorted([x for x in findyaml(self.repodir) if 'activities.yml' in x])

    def _get_geom_yaml(self):
        return sorted([x for x in findyaml(self.repodir) if 'geom.yml' in x])

    def _get_types_yaml(self):
        return sorted([x for x in findyaml(self.repodir) if 'types.yml' in x])

    def _get_fileindex_json(self):
        return sorted([x for x in findjson(self.repodir) if 'file-index.json' in x])

    def _get_activities_json(self):
        return sorted([x for x in findjson(self.repodir) if 'activities.json' in x])
        
    def _get_videos(self):
        return sorted([x for x in findvideo(self.videodir)])
    
    def _parse_video(self, d_videoname_to_path, d_category_to_shortlabel, types_yamlfile, geom_yamlfile, act_yamlfile, stride=1, verbose=False):
        """Reference: https://gitlab.kitware.com/meva/meva-data-repo/-/blob/master/documents/KPF-specification-v4.pdf"""
        
        # Read YAML
        if verbose:
            print('[vipy.dataset.meva.KF1]: Parsing "%s"' % (act_yamlfile))
        geom_yaml = readyaml(geom_yamlfile)
        types_yaml = readyaml(types_yamlfile)
        act_yaml = readyaml(act_yamlfile)

        # Sanity check
        assert act_yamlfile.split('.')[:-2] == geom_yamlfile.split('.')[:-2], "Unmatched activity and geom yaml file"
        assert len(set([types_yaml[0]['meta'], geom_yaml[0]['meta'], act_yaml[0]['meta']]))==1, "Mismatched video name for '%s'" % act_yamlfile
        videoname = act_yaml[0]['meta'] if act_yaml[0]['meta'][-4:] != '.avi' else act_yaml[0]['meta'][0:-4]  # strip .avi 
        if videoname not in d_videoname_to_path:
            if verbose:
                print('[vipy.dataset.meva.KF1]: Invalid MEVA video "%s" in "%s" - Ignoring' % (videoname, filebase(act_yamlfile)))
            return None

        # Parse video
        framerate = 30.0  # All videos are universally 30Hz (from Roddy)
        vid = Scene(filename=d_videoname_to_path[videoname], framerate=framerate)


        # Parse tracks        
        d_id1_to_category = {}
        for t in types_yaml:
            if 'types' in t:
                d_id1_to_category[t['types']['id1']] = list(t['types']['cset3'].keys())[0]

        d_id1_to_track = {}
        d_geom_yaml = groupbyasdict([x['geom'] for x in geom_yaml if 'geom' in x], lambda v: v['id1'])
        assert stride >= 1, "Invalid stride"
        for (id1, geom_yaml) in d_geom_yaml.items():
            geom_yaml = sorted(geom_yaml, key=lambda x: int(x['ts0']))  # increasing
            for (k_geom, v) in enumerate(geom_yaml):
                if stride > 1 and k_geom > 0 and (k_geom < (len(geom_yaml)-stride)) and (k_geom % stride != 0):
                    continue  # Use vipy track interpolation to speed up parsing
                keyframe = int(v['ts0'])
                bb = [int(x) for x in v['g0'].split(' ')]
                bbox = BoundingBox(xmin=bb[0], ymin=bb[1], xmax=bb[2], ymax=bb[3])
                if not bbox.isvalid():
                    if verbose:
                        print('[vipy.dataset.meva.KF1]: Invalid bounding box: id1=%s, bbox="%s", file="%s" - Ignoring' % (str(v['id1']), str(bbox), delpath(self.repodir, geom_yamlfile)))
                elif v['id1'] not in d_id1_to_track:
                    d_id1_to_track[v['id1']] = Track(category=d_id1_to_category[v['id1']], framerate=framerate, keyframes=[keyframe], boxes=[bbox], boundary='strict')
                else:
                    d_id1_to_track[v['id1']].add(keyframe=keyframe, box=bbox)
                
        # Add tracks to scene
        for (k,v) in d_id1_to_track.items():
            try:
                vid.add(v, rangecheck=True)  # throw exception if all tracks are outside the image rectangle
            except Exception as e:
                print('[vipy.dataset.meva.KF1]: track import error "%s" for trackid=%s, track=%s - SKIPPING' % (str(e), k, str(v)))

        # Parse activities
        for v in act_yaml:
            if 'act' in v:
                if 'act2' in v['act']:
                    act2 = v['act']['act2']
                    if isinstance(act2, set):
                        category = list(act2)[0]
                    elif isinstance(act2, dict):
                        category = list(act2.keys())[0]
                    else:
                        raise ValueError('YAML parsing error for "%s"' % str(act2))
                elif 'act3' in v['act']:
                    act3 = v['act']['act3']
                    if isinstance(act3, set):
                        category = list(act3)[0]
                    elif isinstance(act3, dict):
                        category = list(act3.keys())[0]
                    else:
                        raise ValueError('YAML parsing error for "%s"' % str(act3))
                else:
                    raise ValueError('Invalid activity YAML - act2 or act3 must be specified')
                assert len(v['act']['timespan']) == 1, "Multi-span activities not parsed"

                if category not in self.categories():
                    if category in self._d_oldcategory_to_newcategory:
                        category = self._d_oldcategory_to_newcategory[category]  # rationalize
                    else:
                        raise ValueError('undefined category "%s"' % category)
                
                startframe = int(v['act']['timespan'][0]['tsr0'][0])
                endframe = int(v['act']['timespan'][0]['tsr0'][1])
                actorid = [x['id1'] for x in v['act']['actors']]

                for aid in actorid:
                    if not aid in d_id1_to_track:
                        print('[vipy.dataset.meva.KF1]: ActorID %d referenced in activity yaml "%s" not found in geom yaml "%s" - Skipping' % (aid, delpath(self.repodir, act_yamlfile), delpath(self.repodir, geom_yamlfile)))
                
                # Add activity to scene
                tracks = {d_id1_to_track[aid].id():d_id1_to_track[aid] for aid in actorid if aid in d_id1_to_track}
                if len(tracks) > 0:
                    try:
                        vid.add(Activity(category=category, shortlabel=d_category_to_shortlabel[category],
                                         startframe=startframe, endframe=endframe, tracks=tracks, framerate=framerate, attributes={'act':v['act']}), rangecheck=True)
                    except Exception as e:
                        print('[vipy.dataset.meva.KF1]: activity import error "%s" for activity="%s" - SKIPPING' % (str(e), str(v)))
            
        return vid


    def videos(self):
        """Return list of activity videos"""
        return [v for v in self._vidlist if v is not None]

    def tolist(self):
        return self.videos()

    def instances(self, padframes=0):
        """Return list of activity instances"""
        if vipy.globals.num_workers() > 1:
            return [a for A in Batch(self.videos()).activityclip(padframes=padframes) for a in A]
        else:
            warnings.warn('Consider setting vipy.globals.num_workers()>1 to speed this up')
            return [a for v in self.videos() for a in v.activityclip(padframes=padframes)]
        

    def categories(self):
        """Return a list of activity categories"""
        return sorted(list(self._d_category_to_shortlabel.keys()))
        
    def analysis(self, outdir=None):
        """Analyze the MEVA dataset to return helpful statistics and plots"""
        import matplotlib.pyplot as plt        
        
        videos = self._vidlist
        scenes = flatlist([m.activityclip() for m in videos if m is not None])
        activities = flatlist([s.activities().values() for s in scenes])
        tracks = flatlist([s.tracks().values() for s in scenes])
        outdir = tempdir() if outdir is None else outdir
        
        # Category distributions
        d = {}
        d['activity_categories'] = set([a.category() for a in activities])
        d['object_categories'] = set([t.category() for t in tracks])
        d['videos'] = set([v.filename() for v in videos if v is not None])
        d['num_activities'] = sorted([(k,len(v)) for (k,v) in groupbyasdict(activities, lambda a: a.category()).items()], key=lambda x: x[1])
        d['video_density'] = sorted([(v.filename(),len(v.activities())) for v in videos if v is not None], key=lambda x: x[1])
        
        # Histogram of instances
        (categories, freq) = zip(*reversed(d['num_activities']))
        barcolors = ['blue' if not 'vehicle' in c else 'green' for c in categories]
        d['num_activities_histogram'] = vipy.metrics.histogram(freq, categories, barcolors=barcolors, outfile=os.path.join(outdir, 'num_activities_histogram.pdf'), ylabel='Instances')

        # Scatterplot of box sizes
        (x, y, category) = zip(*[(max([t.meanshape()[1] for t in a.tracks().values()]), max([t.meanshape()[0] for t in a.tracks().values()]), a.category()) for a in activities])
        colors = colorlist()
        d_category_to_color = {c:colors[k % len(colors)] for (k,c) in enumerate(category)}
        plt.clf()
        plt.figure()
        plt.subplot(111)
        plt.grid(True)
        d_category_to_boxsizes = groupbyasdict(zip(x,y,category), lambda xyc: xyc[2])
        for c in categories:
            (xc,yc) = zip(*[(xx,yy) for (xx,yy,cc) in d_category_to_boxsizes[c]])
            plt.scatter(xc, yc, c=d_category_to_color[c], label=c)
        plt.xlabel('Bounding box (width)')
        plt.ylabel('Bounding box (height)')
        plt.gca().set_axisbelow(True)                
        plt.axis([0, max(max(x),max(y)), 0, max(max(x),max(y))])
        lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)
        d['bounding_box_scatterplot'] = os.path.join(outdir, 'bounding_box_scatterplot.pdf')
        plt.savefig(d['bounding_box_scatterplot'], bbox_extra_artists=(lgd,), bbox_inches='tight')

        # Separate scatterplots of box sizes
        (x, y, category) = zip(*[(max([t.meanshape()[1] for t in a.tracks().values()]), max([t.meanshape()[0] for t in a.tracks().values()]), a.category()) for a in activities])
        colors = colorlist()
        d_category_to_color = {c:colors[k % len(colors)] for (k,c) in enumerate(category)}
        d_category_to_boxsizes = groupbyasdict(zip(x,y,category), lambda xyc: xyc[2])
        d['category_bounding_box_scatterplot'] = []
        for c in categories:
            plt.clf()
            plt.figure()
            plt.subplot(111)
            plt.grid(True)            
            (xc,yc) = zip(*[(xx,yy) for (xx,yy,cc) in d_category_to_boxsizes[c]])
            plt.scatter(xc, yc, c=d_category_to_color[c], label=c)
            plt.xlabel('Bounding box (width)')
            plt.ylabel('Bounding box (height)')
            plt.legend()            
            plt.gca().set_axisbelow(True)                
            plt.axis([0, max(max(x),max(y)), 0, max(max(x),max(y))])
            filename = os.path.join(outdir, 'category_bounding_box_scatterplot_%s.pdf' % c)
            d['category_bounding_box_scatterplot'].append(filename)
            plt.savefig(filename)
        
        # 2D histogram of box sixes
        plt.clf()
        plt.figure()
        plt.hist2d(x, y, bins=10)
        plt.xlabel('Bounding box (width)')
        plt.ylabel('Bounding box (height)')        
        d['2D_bounding_box_histogram'] = os.path.join(outdir, '2D_bounding_box_histogram.pdf')
        plt.savefig(d['2D_bounding_box_histogram'])
        
        # Scatterplot of people and vehicles box sizes
        plt.clf()
        plt.figure()
        plt.grid(True)
        d_category_to_color = {'person':'blue', 'vehicle':'green'}
        for c in ['person', 'vehicle']:
            (xc, yc) = zip(*[(t.meanshape()[1], t.meanshape()[0]) for t in tracks if t.category() == c])
            plt.scatter(xc, yc, c=d_category_to_color[c], label=c)
        plt.xlabel('bounding box (width)')
        plt.ylabel('bounding box (height)')
        plt.axis([0, max(max(x),max(y)), 0, max(max(x),max(y))])                
        plt.legend()
        plt.gca().set_axisbelow(True)        
        d['object_bounding_box_scatterplot'] = os.path.join(outdir, 'object_bounding_box_scatterplot.pdf')
        plt.savefig(d['object_bounding_box_scatterplot'])
        
        # 2D histogram of people and vehicles box sizes
        for c in ['person', 'vehicle']:
            (x, y) = zip(*[(t.meanshape()[1], t.meanshape()[0]) for t in tracks if t.category() == c])
            plt.clf()
            plt.figure()
            plt.hist2d(x, y, bins=10)
            plt.xlabel('Bounding box (width)')
            plt.ylabel('Bounding box (height)')
            
            d['2D_%s_bounding_box_histogram' % c] = os.path.join(outdir, '2D_%s_bounding_box_histogram.pdf' % c)
            plt.savefig(d['2D_%s_bounding_box_histogram' % c])

        return d



import os
import sys
from vipy.util import try_import, islist, tolist, tempdir, remkdir
from itertools import repeat
try_import('dask', 'dask distributed torch')
from dask.distributed import as_completed, wait
try_import('torch', 'torch');  import torch
import numpy as np
import tempfile
import warnings
import vipy.globals


class Batch(object):
    """vipy.batch.Batch class

    This class provides a representation of a set of vipy objects.  All of the object types must be the same.  If so, then an operation on the batch is performed on each of the elements in the batch in parallel.

    Examples:

    >>> b = vipy.batch.Batch([Image(filename='img_%06d.png' % k) for k in range(0,100)])
    >>> b.bgr()  # convert all elements in batch to BGR
    >>> b.torch()  # load all elements in batch and convert to torch tensor
    >>> b.map(lambda im: im.bgr())  # equivalent
    >>> b.map(lambda im: np.sum(im.array())) 
    >>> b.map(lambda im, f: im.saveas(f), args=['out%d.jpg' % k for k in range(0,100)])
    
    >>> v = vipy.video.RandomSceneActivity()
    >>> b = vipy.batch.Batch(v, n_processes=16)
    >>> b.map(lambda v,k: v[k], args=[(k,) for k in range(0, len(v))])  # paralle interpolation

    >>> d = vipy.dataset.kinetics.Kinetics700('/path/to/kinetics').download().trainset()
    >>> b = vipy.batch.Batch(d, n_processes=32)
    >>> b.map(lambda v: v.download().save())  # will download and clip dataset in parallel

    """    
             
    def __init__(self, objlist, n_processes=2, dashboard=False):
        """Create a batch of homogeneous vipy.image objects from an iterable that can be operated on with a single parallel function call
        """
        objlist = tolist(objlist)
        self._batchtype = type(objlist[0])        
        assert all([isinstance(im, self._batchtype) for im in objlist]), "Invalid input - Must be homogeneous list of the same type"                
        self._objlist = objlist        
        if vipy.globals.dask() is None or vipy.globals.num_workers() < n_processes:
            vipy.globals.dask(num_processes=n_processes, dashboard=dashboard)
        self._client = vipy.globals.dask().client()  # shutdown using vipy.globals.dask().shutdown(), or let python garbage collect it

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return self
    
    def __len__(self):
        return len(self._objlist)

    def __repr__(self):
        return str('<vipy.batch: type=%s, len=%d, procs=%d>' % (str(self._batchtype), len(self), self.n_processes()))

    def info(self):
        return self._client.scheduler_info()

    def n_processes(self):
        return len(self.info()['workers'])

    def batch(self, newlist=None):
        if islist(newlist) and not hasattr(newlist[0], 'result'):
            self._objlist = newlist
            return self
        elif islist(newlist) and hasattr(newlist[0], 'result'):
            try:
                wait(newlist)
            except KeyboardInterrupt:
                warnings.warn('[vipy.batch]: batch cannot be restarted after killing with ctrl-c - You must create a new Batch()')
                vipy.globals.dask().shutdown()
                self._client = None
                return None  
            except:
                # warnings.warn('[vipy.batch]: batch cannot be restarted after exception - Recreate Batch()')                
                raise
            completedlist = [f.result() for f in newlist]
            if isinstance(completedlist[0], self._batchtype):
                self._objlist = completedlist
                return self
            else:
                return completedlist
        elif newlist is None:
            return self._objlist
        else:
            raise ValueError('Invalid input - must be list')
        
    def __iter__(self):
        for im in self._objlist:
            yield im
            
    def __getattr__(self, attr):
        """Call the same method on all Image objects"""
        assert self.__dict__['_client'] is not None, "Batch() must be reconstructed after shutdown"                                
        return lambda *args, **kw: self.batch(self.__dict__['_client'].map(lambda im: getattr(im, attr)(*args, **kw), self._objlist))

    def product(self, f_lambda, args, waiting=True):
        """Cartesian product of args and batch, returns an MxN list of N args applied to M batch elements.  Use this with extreme caution, as the memory requirements may be high."""
        assert self.__dict__['_client'] is not None, "Batch() must be reconstructed after shutdown"                        
        c = self.__dict__['_client']
        objlist = c.scatter(self._objlist)        
        futures = [c.submit(f_lambda, im, *a) for im in objlist for a in args]
        return self.batch(futures) if waiting else futures
        
    def map(self, f_lambda, args=None):
        """Run the lambda function on each of the elements of the batch. 
        
        If args is provided, then this is a unique argument for the lambda function for each of the elements in the batch, or is broadcastable.
        
        >>> iml = [vipy.image.RandomScene(512,512) for k in range(0,1000)]   
        >>> imb = vipy.image.Batch(iml, n_processes=4) 
        >>> imb.map(lambda im,f: im.saveas(f), args=[('/tmp/out%d.jpg'%k,) for k in range(0,1000)])  
        >>> imb.map(lambda im: im.rgb())  # this is equivalent to imb.rgb()

        """
        assert self.__dict__['_client'] is not None, "Batch() must be reconstructed after shutdown"                
        c = self.__dict__['_client']        
        if args is not None:
            if len(self._objlist) > 1:
                assert islist(args) and len(list(args)) == len(self._objlist), "args must be a list of arguments of length %d, one for each element in batch" % len(self._objlist)
                objlist = c.scatter(self._objlist)
                return self.batch([c.submit(f_lambda, im, *a) for (im, a) in zip(objlist, args)])                
            else:
                assert islist(args), "args must be a list"
                obj = c.scatter(self._objlist[0], broadcast=True)
                return self.batch([self.__dict__['_client'].submit(f_lambda, obj, *a) for a in args])
        else:
            return self.batch(self.__dict__['_client'].map(f_lambda, self._objlist))

    def filter(self, f_lambda):
        """Run the lambda function on each of the elements of the batch and filter based on the provided lambda  
        """
        assert self.__dict__['_client'] is not None, "Batch() must be reconstructed after shutdown"        
        c = self.__dict__['_client']
        objlist = c.scatter(self._objlist)        
        is_filtered = self.batch(self.__dict__['_client'].map(f_lambda, objlist))
        self._objlist = [obj for (f, obj) in zip(is_filtered, self._objlist) if f is True]
        return self
        
    def torch(self):
        """Convert the batch of N HxWxC images to a NxCxHxW torch tensor"""
        return torch.cat(self.map(lambda im: im.torch()))

    def numpy(self):
        """Convert the batch of N HxWxC images to a NxCxHxW torch tensor"""
        return np.stack(self.map(lambda im: im.numpy()))
    



from setuptools import setup, find_packages

d_version = {}
with open("./vipy/version.py") as fp:
    exec(fp.read(), d_version)
version = d_version['VERSION']

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='vipy',
    author='Jeffrey Byrne',
    author_email='jeff@visym.com',
    version=version,
    packages=find_packages(),
    description='Visym Python Tools for Privacy Preserving Computer Vision',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/visym/vipy',
    download_url='https://github.com/visym/vipy/archive/%s.tar.gz' % version,
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "dill",
        "pillow",
        "ffmpeg-python"
    ],
    keywords=['vision', 'learning', 'ML', 'CV'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License"
    ]
)



import os
import numpy as np
import vipy.video
import vipy.videosearch
import vipy.object
from vipy.util import tempjpg, tempdir, Failed, isurl, rmdir, totempdir
from vipy.geometry import BoundingBox
import pdb
from vipy.dataset.kinetics import Kinetics400, Kinetics600, Kinetics700
from vipy.dataset.activitynet import ActivityNet
from vipy.dataset.lfw import LFW
from vipy.object import Detection, Track, Activity
import shutil


mp4file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Video.mp4')
mp4url = 'https://www.youtube.com/watch?v=C0DPdy98e4c'

def test_video():
    # Common Parameters
    urls = vipy.videosearch.youtube('owl',1)
    if len(urls) > 0:
        assert isurl(urls[0])
    print('[test_video.video]: videosearch   PASSED')
    
    # Empty constructor
    try:
        v = vipy.video.Video()
        raise Failed()
    except Failed:
        raise
    except:
        print('[test_video.video]: Empty constructor  PASSED')

    # Malformed URL should raise exception
    try:
        v = vipy.video.Video(url='myurl')
        raise Failed()
    except Failed:
        raise
    except:
        print('[test_video.video]: Malformed URL constructor   PASSED')

    # Invalid constructors
    try:
        v = vipy.video.Video(array=np.random.rand(1,10,10,3))
        raise Failed()  # np.float64 not allowed
    except Failed:
        raise
    except:
        pass
    try:
        v = vipy.video.Video(array=np.random.rand(1,10,10,3).astype(np.float32), colorspace='rgb')
        raise Failed()  # rgb must be uint8
    except Failed:
        raise
    except:
        pass
    try:
        v = vipy.video.Video(array=np.random.rand(1,10,10,2).astype(np.uint8), colorspace='rgb')
        raise Failed()  # rgb must be uint8, three channel
    except Failed:
        raise
    except:
        pass
    try:
        v = vipy.video.Video(array=np.random.rand(1,10,10,3).astype(np.uint8), colorspace='lum')
        raise Failed()  # lum must be one channel
    except Failed:
        raise
    except:
        pass

    # Valid URL should not raise exception (even if it is not an image extension)
    im = vipy.video.Video(url='http://visym.com')
    print('[test_video.video]: Image URL constructor  PASSED')
        
    # Valid constructors
    v = vipy.video.Video(url=mp4url)
    v.__repr__()    
    v = vipy.video.Video(filename=mp4file)
    v.__repr__()
    v = vipy.video.Video(filename=mp4file, url=mp4url)
    v.__repr__()
    v = vipy.video.Video(array=np.random.rand(1,10,10,5).astype(np.float32), colorspace='float')
    v.__repr__()
    v = vipy.video.Video(array=np.random.rand(1,10,10,3).astype(np.uint8), colorspace='rgb')
    v.__repr__()
    v = vipy.video.Video(array=np.random.rand(5,10,10,3).astype(np.uint8), colorspace='bgr')
    v.__repr__()
    v = vipy.video.Video(array=np.random.rand(5,10,10,1).astype(np.uint8), colorspace='lum')
    v.__repr__()
    print('[test_video.video]: __repr__  PASSED')    
    

    # Saveas
    v = vipy.video.Video(url=mp4url)
    v.saveas(ignoreErrors=True)
    print('[test_video.video]: saveas(ignoreErrors=True)  PASSED')        

    
    # Clone
    v = vipy.video.RandomVideo(64,64,64)
    vc = v.clone()
    assert np.allclose(vc._array, v._array)    
    vc._array = 0
    assert not np.allclose(vc._array, v._array)
    vc = v.clone(flushforward=True)
    assert vc._array is None and v._array is not None
    vc = v.clone(flushbackward=True)
    assert vc._array is not None and v._array is None
    vc = v.clone(flush=True)
    assert vc._array is None and v._array is None
    print('[test_video.scene]: clone()  PASSED')        

    
def _test_scene():

    # Activityclip
    v = vipy.video.RandomSceneActivity(64,64,64)
    vc = v.clone(flushforward=True).filename('Video.mp4')
    assert vc._array is None and v._array is not None
    activitylength = [len(a) for a in vc.activities().values()]
    assert all([len(c.activities())==1 for c in vc.activityclip()])    
    assert all([len(a)==al for (c,al) in zip(vc.activityclip(padframes=0), activitylength) for a in c.activities().values()])
    try:
        vc.activityclip(padframes=2)  # will result in startframe < 0
        Failed()
    except Failed:
        raise
    except:
        pass
    print('[test_video.scene]: activityclip()  PASSED')    
    
    # Activitycrop
    v = vipy.video.RandomSceneActivity(64,64,64)
    vc = v.clone(flushforward=True).filename('Video.mp4')
    assert vc._array is None and v._array is not None
    a = vc.activitytube()
    print('[test_video.scene]: activitytube()  PASSED - basic only')        
    
    # Downloader
    v = vipy.video.Video(url='http://visym.com/out.mp4').load(ignoreErrors=True)
    print('[test_video.video]: download ignoreErrors  PASSED')                
    v = vipy.video.Scene(url=mp4url).clip(0,100).load()
    print('[test_video.scene: download  PASSED')        
    for im in v:
        assert im.shape() == v.shape()
    print('[test_video.scene]: __iter__  PASSED')
    
    
    vid = vipy.video.Scene(filename=mp4file, tracks=[vipy.object.Track(category='person', keyframes=[0,200], boxes=[BoundingBox(xmin=0,ymin=0,width=200,height=400), BoundingBox(xmin=0,ymin=0,width=400,height=100)]),
                                                     vipy.object.Track(category='vehicle', keyframes=[0,200], boxes=[BoundingBox(xmin=100,ymin=200,width=300,height=400), BoundingBox(xmin=400,ymin=300,width=200,height=100)])])

    # Loader
    try:
        v = vid.clone().load(startframe=0)
        Failed()
    except Failed:
        raise
    except:
        pass
    try:
        v = vid.clone().load(endframe=0)
        Failed()
    except Failed:
        raise
    except:
        pass
    try:
        v = vid.clone().load(mindim=1, rescale=1)
        Failed()
    except Failed:
        raise
    except:
        pass
    try:
        v = vid.clone().load(rotation='an_invalid_string')
        Failed()
    except Failed:
        raise
    except:
        pass

    v = vid.clone().load(startframe=0, endframe=10)
    assert len(v) == 10
    assert np.allclose(vid.clone().load(startframe=0, endframe=10, rotation='rot90ccw').array(), vid.clone().clip(0,10).rot90ccw().load().array())
    assert np.allclose(vid.clone().load(startframe=1, endframe=11, rotation='rot90cw').array(), vid.clone().clip(1,11).rot90cw().load().array())
    assert np.allclose(vid.clone().load(startframe=2, endframe=12, mindim=10).array(), vid.clone().clip(2,12).mindim(10).load().array())
    assert np.allclose(vid.clone().load(startframe=3, endframe=12, rescale=0.2).array(), vid.clone().clip(3,12).rescale(0.2).load().array())
    print('[test_video.scene]:  load: PASSED')
    
    v = vid.clone().clip(0,10).load()
    assert len(v) == 10
    vc = v.clone(flushforward=True).clip(1,4).load()
    assert len(vc) == 3
    print('[test_video.scene]:  trim: PASSED')

    (h,w) = v.shape()
    v = vid.clone().clip(0,10).rot90ccw().load()
    assert v.width() == h and v.height() == w and len(v) == 10
    assert [im.crop().height() for im in v[0]] == [200, 300]
    print('[test_video.scene]: rot90ccw  PASSED')

    v = vid.clone().clip(0,10).rot90cw().load()
    assert v.width() == h and v.height() == w and len(v) == 10
    assert [im.crop().height() for im in v[0]] == [200, 300]    
    print('[test_video.scene]: rot90cw  PASSED')

    v = vid.clone().clip(0,10).rescale(0.5).load(verbose=False)
    assert v.height() * 2 == h and v.width() * 2 == w and len(v) == 10
    assert [im.crop().height() for im in v[0]] == [200, 200]
    assert [im.crop().width() for im in v[0]] == [100, 150]        
    print('[test_video.scene]: rescale  PASSED')

    (H,W) = vid.clone().clip(0,200).load(verbose=False).shape()
    v = vid.clone().clip(0,200).resize(cols=100).load(verbose=False)    
    assert v.width() == 100 and len(v) == 200
    assert np.allclose([im.bbox.height() for im in v[0]], [400*(100.0/W), 400*(100.0/W)])
    assert np.allclose([im.bbox.width() for im in v[0]], [200*(100.0/W), 300*(100.0/W)])
    print('[test_video.scene]: resize isotropic  PASSED')
    
    v = vid.clone().resize(cols=100, rows=100).clip(0,11).load(verbose=False)    
    assert v.width() == 100 and v.height() == 100 and len(v) == 11
    assert np.allclose([im.bbox.height() for im in v[0]], [400*(100.0/H), 400*(100.0/H)])
    assert np.allclose([im.bbox.width() for im in v[0]], [200*(100.0/W), 300*(100.0/W)])    
    print('[test_video.scene]: resize anisotropic   PASSED')
    
    v = vid.clone().clip(0,200).rot90cw().resize(rows=200).load(verbose=False)
    assert v.height() == 200 and len(v) == 200
    print('[test_video.scene]: rotate and resize  PASSED')
    
    v.annotate('vipy.mp4')
    print('[test_video.scene]: annotate  PASSED')

    v = vid.clone().clip(150,200).rot90cw().resize(rows=200).load(verbose=False)
    assert v.height() == 200 and len(v) == 50
    print('[test_video.scene]: trim, rotate, resize  PASSED')

    v = vid.clone().crop(BoundingBox(xmin=1, ymin=2, width=10, height=20)).load(verbose=False)
    assert v.height() == 20 
    assert v[0][0].bbox.xmin() == -1 and v[0][0].bbox.ymin() == -2
    assert v[0][1].bbox.xmin() == 99 and v[0][1].bbox.ymin() == 198
    print('[test_video.scene]: crop  PASSED')

    # Bounding box interpolation
    for im in v:
        assert im.shape() == v.shape()
    assert len(im) == 0  # strict boundary by default
    assert v[200][0].bbox.width() == 400 and v[200][1].bbox.width() == 200
    print('[test_video.scene]: frame interpolation  PASSED')

    # Thumbnail
    v = vid.clone().thumbnail()
    assert os.path.exists(v)
    print('[test_video.scene]: thumbnail ("%s")  PASSED' % v)

    # Map
    v = vid.clone().clip(0,10).map(lambda img: img*0)
    assert (np.sum(v.array().flatten()) == 0)
    print('[test_video.scene]: map ("%s")  PASSED' % v)
    
    # Shared array
    img = np.random.rand(2,2,2,3).astype(np.float32)
    v = vipy.video.Video(array=img)
    img[0,:,:,:] = 0
    assert np.sum(v.array()[0]) == 0
    img = np.random.rand(2,2,2,3).astype(np.float32)    
    v = v.array(img, copy=False)
    img[0,:,:,:] = 0    
    assert np.sum(v.array()[0]) == 0
    img = np.random.rand(2,2,2,3).astype(np.float32)        
    v = v.fromarray(img)
    img[0,:,:,:] = 0
    assert np.sum(v.array()[0]) != 0    
    print('[test_video.scene]: array by reference  PASSED')

    # Mutable iterator
    frames = np.random.rand(2,2,2,3).astype(np.float32)
    v = vipy.video.Video(array=frames)
    for im in v:
        im.numpy()[:,:] = 0
    assert np.sum(v.array().flatten()) == 0
    frames = np.random.rand(2,2,2,3).astype(np.float32)
    v = vipy.video.Video(array=frames)
    for im in v.numpy():
        im[:,:] = 0
    assert np.sum(v.array().flatten()) == 0
    v = vipy.video.Video(mp4file).clip(0,10).load()
    for im in v.numpy():
        im[:,:] = 0
    assert np.sum(v.array().flatten()) == 0
    v = vipy.video.Video(mp4file).clip(0,10).load()
    for im in v:
        img = im.numpy()
        img[:,:] = 0
    assert np.sum(v.array().flatten()) == 0
    print('[test_video.scene]: mutable iterator  PASSED')    

    # Scene iterator
    frames = np.random.rand(2,2,2,3).astype(np.float32)
    v = vipy.video.Scene(array=frames)
    for im in v:
        v.add(Detection(0, 0, 0, 100, 100))
        v.add(Track(category=1, keyframes=[1], boxes=[BoundingBox(0,0,1,1)]))
        v.add([1,2,3,4], category='test', rangecheck=False)
    print('[test_video.scene]: scene iterator  PASSED')
    
    # Random scenes
    v = vipy.video.RandomVideo(64,64,64)
    v = vipy.video.RandomScene(64,64,64)
    vorig = vipy.video.RandomSceneActivity(64,64,64)
    print(vorig[0])
    print(vorig[0][0])
    print('[test_video.scene]: random scene  PASSED')
    
    # Video scenes
    v1 = vipy.video.Video('Video.mp4', startframe=50, endframe=100).load()
    v2 = vipy.video.Video('Video.mp4').clip(0,100).load()    
    assert v1[0] == v2[50]
    print('[test_video.scene]: video scenes  PASSED')    


    # Saveas
    outfile = totempdir('Video.mp4')
    shutil.copy('Video.mp4', outfile)
    v = vipy.video.Video(filename=outfile, startframe=0, endframe=10)
    v.saveas()
    assert v.hasfilename() and os.path.getsize('Video.mp4') != os.path.getsize(v.filename())
    print('[test_video.scene]: saveas()  PASSED')    


def test_track():
    t = Track(category=1, keyframes=[1,10], boxes=[BoundingBox(0,0,10,11), BoundingBox(1,1,11,12)])
    assert t.boundingbox().xywh() == (0,0,11,12)
    print('[test_video.track]: boundingbox()  PASSED')
    
    assert t.during(1) and t.during(10) and t.during(5)
    assert not t.during(11)
    print('[test_video.track]: during()  PASSED')    

    t.dict()
    print('[test_video.track]: dict()  PASSED')

    for d in t:
        assert d.width() == 10 and d.height() == 11
    assert d.xmin() == 1 and d.ymin() == 1
    print('[test_video.track]: interpolation  PASSED')

    
def test_scene_union():
    
    vid = vipy.video.RandomVideo(64,64,32)
    (rows, cols) = vid.shape()
    track1 = vipy.object.Track(label='Person1', keyframes=[0], boxes=vipy.geometry.BoundingBox(10,20,30,40)).add(2, vipy.geometry.BoundingBox(20,30,40,50))
    track2 = vipy.object.Track(label='Person2', keyframes=1, boxes=[vipy.geometry.BoundingBox(10,20,30,40)]).add(3, vipy.geometry.BoundingBox(30,40,50,60))
    activity = vipy.object.Activity(label='act1', startframe=0, endframe=3).add(track1).add(track2)

    assert track1.boundingbox().ulbr() == (10,20,40,50)
    assert track2.boundingbox().ulbr() == (10,20,50,60)
    assert activity.boundingbox().ulbr() == vipy.geometry.BoundingBox(10,20,50,60).ulbr()
    assert activity[0][0].ulbr() == vipy.geometry.BoundingBox(10,20,30,40).ulbr()
    assert activity[3][0].ulbr() == vipy.geometry.BoundingBox(30,40,50,60).ulbr()                
    assert activity[1][0].ulbr() == track1[1].ulbr()
    assert activity[1][1].ulbr() == track2[1].ulbr()
    
    # By reference
    assert activity.tracks()[track1.id()].category() == 'Person1'
    track1.category('PersonA')
    assert activity.tracks()[track1.id()].category() == 'PersonA'    
    assert 'PersonA' in activity.categories() 
    track1.category('Person1')
    
    v = vipy.video.Scene(array=vid.array(), colorspace='rgb', category='scene', activities=[activity])
    vu = v.clone().union(v)
    assert len(vu.activities())==1

    activity = vipy.object.Activity(label='act2', startframe=0, endframe=3).add(track1).add(track2)
    v2 = vipy.video.Scene(array=v.array(), colorspace='rgb', category='scene', activities=[activity])
    vu = v.clone().union(v2)
    assert len(vu.activities())==2
    assert vu.categories() == set(['act2', 'act1', 'Person1', 'Person2'])
    
    track3 = vipy.object.Track(label='Person3', keyframes=[1], boxes=[vipy.geometry.BoundingBox(10,20,30,40)]).add(3, vipy.geometry.BoundingBox(30,40,50,60))    
    activity = vipy.object.Activity(label='act1', startframe=0, endframe=2).add(track1).add(track3)
    v2 = vipy.video.Scene(array=vid.array(), colorspace='rgb', category='scene', activities=[activity])
    vu = v.clone().union(v2)
    assert len(vu.categories()) == 4

    track3 = vipy.object.Track(label='Person3', keyframes=[0], boxes=[vipy.geometry.BoundingBox(10,20,30,40)]).add(2, vipy.geometry.BoundingBox(20,30,40,100))     
    activity = vipy.object.Activity(label='act2', startframe=0, endframe=2).add(track1).add(track3)
    v2 = vipy.video.Scene(array=v.array(), colorspace='rgb', category='scene', activities=[activity])
    vu = v.clone().union(v2)
    
    assert len(vu.categories()) == 5
    print('[test_video.track]: test_scene_union  PASSED')
    
    
if __name__ == "__main__":
    test_video()
    test_track()
    _test_scene()
    test_scene_union()



import os
import sys


def test_import():
    import vipy
    import vipy.activity
    import vipy.annotation
    import vipy.calibration
    import vipy.camera
    import vipy.collector
    import vipy.downloader
    import vipy.dropbox
    import vipy.geometry
    import vipy.image
    import vipy.linalg
    import vipy.math
    import vipy.metrics
    import vipy.object
    import vipy.show
    import vipy.ssim
    import vipy.util
    import vipy.video
    import vipy.videosearch
    import vipy.visualize
    import vipy.gui.using_matplotlib
    import vipy.dataset.aflw
    import vipy.dataset.caltech256
    import vipy.dataset.ethzshapes
    import vipy.dataset.fddb
    import vipy.dataset.imagenet
    import vipy.dataset.kthactions
    import vipy.dataset.megaface
    import vipy.dataset.msceleb
    import vipy.dataset.vggface
    import vipy.dataset.youtubefaces
    import vipy.dataset.activitynet
    import vipy.dataset.caltech101
    import vipy.dataset.casia
    import vipy.dataset.facescrub
    import vipy.dataset.hmdb
    import vipy.dataset.kinetics
    import vipy.dataset.lfw
    import vipy.dataset.mnist
    import vipy.dataset.vggface2
    print('[test_import]: PASSED')


if __name__ == "__main__":
    test_import()
